/*
 * UserDriverLCD.c -  LCD UserDriver for SilentC
 *
 * 2008.11.01 by masato
 *
 *
 * PORTUB 0 to 3 : LCD DB 4 to 7
 * PORTUA 2 : LCD RS
 * POERUA 3 : LCD E
 * DTOUT0 : Exciter
 */

#include "support_common.h"

/* LCD commands */

#define LCD_CMD_RS                      0
#define LCD_DATA_RS                     1

#define LCD_CMD_INT8B           0x03
#define LCD_CMD_INT4B           0x02
#define LCD_CMD_CLAR            0x01
#define LCD_CMD_HOME            0x02
#define LCD_CMD_FUNC_SET        0x28 /* 8bit, 2line, 5*8dots */ 
#define LCD_CMD_ENT_SET         0x06
#define LCD_CMD_DISP_CUR        0x0e
#define LCD_CMD_DISP_BNK        0x0d
#define LCD_CMD_DISP_NCUR       0x0c
#define LCD_CMD_DISP_OFF        0x08

#define LCD_COLS        16

/* jump table */

void LCD_Init();
void LCD_Test();
void LCD_L1Out(unsigned char *p);
void LCD_L2Out(unsigned char *p);

asm void jump_table(void) {
  jmp LCD_Init
  jmp LCD_L1Out
  jmp LCD_L2Out
  jmp LCD_Test
}

/*
** ===================================================================
**     Method      :  Cpu_Delay100US (bean MCF52233_80_LQFP)
**
**     Description :
**         This method realizes software delay. The length of delay is
**         at least 100 microsecond multiply input parameter [us100].
**         As the delay implementation is not based on real clock, the
**         delay time may be increased by interrupt service routines
**         processed during the delay. The method is independent on
**         selected speed mode.
**     Parameters  :
**         NAME            - DESCRIPTION
**         us100           - Number of 100 us delay repetitions.
**     Returns     : Nothing
** ===================================================================
*/
__declspec(register_abi) void Cpu_Delay100US(uint16 us100:__D0)
{
  /* irremovable one time overhead (ignored): 10 cycles */
  /* move: 1 cycle overhead (load parameter into register) */
  /* jsr:  3 cycles overhead (jump to subroutine) */
  /* andi: 1 cycle overhead (clear upper word of d0) */
  /* rts:  5 cycles overhead (return from subroutine) */

  /* aproximate irremovable overhead for each 100us cycle (counted) : 3 cycles */
  /* subq.l 1 cycles overhead  */
  /* bne.b  2 cycles overhead  */

#pragma unused(us100)
  asm {
    naked
    andi.l #65535,d0                   /* parameter is word - clear the rest of d0 register */
loop:
    /* 100 us delay block begin */
    /*
     * Delay
     *   - requested                  : 100 us @ 60MHz,
     *   - possible                   : 6000 c, 100000 ns
     *   - without removable overhead : 5997 c, 99950 ns
     */
    move.l #1998,d1                    /* (1 c: 16.67 ns) number of iterations */
label0:
    subq.l #1,d1                       /* (1 c: 16.67 ns) decrement d1 */
    bne.b label0                       /* (2 c: 33.33 ns) repeat 1998x */
    tpf                                /* (1 c: 16.67 ns) wait for 1 c */
    /* 100 us delay block end */
    subq.l #1,d0                       /* parameter is passed via d0 register */
    bne.b loop                         /* next loop */
    rts                                /* return from subroutine */
  }
}

/*
 * LCD_Exciter 100KHz PWM out on DOUT0
 */
void LCD_Exciter(void)
{
  /* PTCPAR: PTCPAR0=2 */
  MCF_GPIO_PTCPAR = (uint8)((MCF_GPIO_PTCPAR & (~(uint8)(1))) | (uint8)(2));
 
  /* DTMR0: PS=0,CE=0,OM=1,ORRI=0,FRR=1,CLK=1,RST=1 */
  MCF_DTIM0_DTMR = (uint16)41;
  /* DTXMR0: DMAEN=0,HALTED=0,??=0,??=0,??=0,??=0,??=0,MODE16=0 */
  MCF_DTIM0_DTXMR = (uint16)0;
  /* DTER0: ??=0,??=0,??=0,??=0,??=0,??=0,REF=1,CAP=0 */
  MCF_DTIM0_DTER = (uint8)2;
  /* DTRR0: REF=599 */
  MCF_DTIM0_DTRR = (uint32)599;
  /* DTCN0: CNT=0 */
  MCF_DTIM0_DTCN = (uint32)0;
  /* Run Counter */
  MCF_DTIM0_DTMR = (uint16)((MCF_DTIM0_DTMR & ~0x06) | 0x02);
}

/*
 * LCD_DB4Bits_PutVal - output PORTUB 0 to 3
 */
void LCD_DB4Bits_PutVal(uint8 Val)
{
  Val &= 15;                           /* Mask output value */
  MCF_GPIO_PORTUB = (uint8)((MCF_GPIO_PORTUB & (~15)) | Val);
}

/*
 * LCD_DB4Bits_Init - initialize PORTUB (0 to 3) and PORTUA (2 and 3)
 */
void LCD_DB4Bits_Init(void)
{
  // PORTUB 0..3
  MCF_GPIO_PORTUB = (uint8)(MCF_GPIO_PORTUB & ~15);
  // DDRUB 0..3 to output
  MCF_GPIO_DDRUB = 15;
  // PUBPAR 0..3
  MCF_GPIO_PUBPAR = 0;
  
  // PORTUA 2,3
  MCF_GPIO_PORTUA = (uint8)(MCF_GPIO_PORTUA & ~12);
  // DDRUA
  MCF_GPIO_DDRUA = 12;
  // PUAPAR
  MCF_GPIO_PUAPAR = (uint8)(MCF_GPIO_PUAPAR & ~240);
}

/*
 * LCD_RS_PutVal - LCD RS
 */
void LCD_RS_PutVal(unsigned char Val)
{
  if (Val) {
    MCF_GPIO_SETUA = 4;                 /* SETUA2=1 */
  } else { /* !Val */
    MCF_GPIO_CLRUA = 251;               /* CLRUA2=0 */
  }
}

/*
 * LCD_E_PutVal - LCD E
 */
void LCD_E_PutVal(unsigned char Val)
{
  if (Val) {
    MCF_GPIO_SETUA = 8;                 /* SETUA3=1 */
  } else { /* !Val */
    MCF_GPIO_CLRUA = 247;               /* CLRUA3=0 */
  } /* !Val */

}

/*
 * LCD_StrobePulse
 */
void LCD_StrobePulse(void)
{
    LCD_E_PutVal(1);
    Cpu_Delay100US(1);
    LCD_E_PutVal(0);
    Cpu_Delay100US(1);
}

/*
 * LCD_WriteData (4 bit mode)
 */
void LCD_WriteData(unsigned char data, unsigned char rs)
{
  unsigned char   upper_data;
  upper_data      = (uint8)(data >> 4);
        
  LCD_RS_PutVal(rs);
  LCD_E_PutVal(0);
  LCD_DB4Bits_PutVal(upper_data);
  LCD_StrobePulse();
  LCD_DB4Bits_PutVal(data);
  LCD_StrobePulse();
}

void LCD_Init(void)
{
  LCD_Exciter();
  LCD_DB4Bits_Init();
        
  LCD_RS_PutVal(LCD_CMD_RS);              /* init E and RS */
  LCD_E_PutVal(0);
  Cpu_Delay100US(1000);                   /* wait 100msec */
        
  LCD_DB4Bits_PutVal(LCD_CMD_INT8B);      /* set 8 bit mode */
  LCD_StrobePulse();                         
  Cpu_Delay100US(150);
  LCD_DB4Bits_PutVal(LCD_CMD_INT8B);
  LCD_StrobePulse();
  Cpu_Delay100US(3);
  LCD_DB4Bits_PutVal(LCD_CMD_INT8B);
  LCD_StrobePulse();
  Cpu_Delay100US(1);
  LCD_DB4Bits_PutVal(LCD_CMD_INT4B);     /* 4 bit mode */
  LCD_StrobePulse();
  Cpu_Delay100US(1);
 
  LCD_WriteData(LCD_CMD_FUNC_SET, LCD_CMD_RS);  /* 0x38: 8BitMODE 02 lines */
  Cpu_Delay100US(1);
  LCD_WriteData(LCD_CMD_DISP_OFF, LCD_CMD_RS);  /* 0x08: display off */
  Cpu_Delay100US(1);
  LCD_WriteData(LCD_CMD_CLAR, LCD_CMD_RS);      /* 0x01: clear display */
  Cpu_Delay100US(50);
  LCD_WriteData(LCD_CMD_ENT_SET,  LCD_CMD_RS);  /* 0x06: entry mode */
  Cpu_Delay100US(1);
  LCD_WriteData(LCD_CMD_DISP_NCUR,LCD_CMD_RS);	/* 0x0C: display on */
  Cpu_Delay100US(1);
}

/*
 * LCD_putc - write one char to LCD
 */
void LCD_putc(unsigned char c)
{
  LCD_WriteData(c, LCD_DATA_RS);		/* write to VRAM */
  Cpu_Delay100US(1);                            /* wait 100us */
}

/* 
 * LCD_PrintDisplay - write one line to LCD
 */
uint32 LCD_PrintDisplay(uint32 LineAddr, unsigned char* s)
{
  int i;
  unsigned char *str, c;
  str = s;
  LCD_WriteData((uint8)LineAddr,  LCD_CMD_RS);

  i = 0;
  while (*str != '\0') {
    c = *str++;
    if (c < ' ') c = ' ';
    LCD_WriteData(c, LCD_DATA_RS);
    i++;
    if (i > LCD_COLS) break;
  }
  for ( ; i < LCD_COLS; i++)
    LCD_WriteData(' ', LCD_DATA_RS);
  return LineAddr;
}

void LCD_Test()
{
  int i;

  LCD_Init();

  for(i = 0; i < 1; i++) {
    LCD_PrintDisplay(0x80, (unsigned char*)"CQ-FRK-MCF52233");
    LCD_PrintDisplay(0xC0, (unsigned char*)"Hello, World!");
  }
}

void LCD_L1Out(unsigned char *p)
{
  LCD_PrintDisplay(0x80, p);
}

void LCD_L2Out(unsigned char *p)
{
  LCD_PrintDisplay(0xC0, p);
}

/* EOF */
